package dev.fitko.fitconnect.cli;

import com.nimbusds.jose.jwk.JWK;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.exceptions.internal.BatchImportException;
import dev.fitko.fitconnect.cli.batch.BatchImporter;
import dev.fitko.fitconnect.cli.batch.ImportRecord;
import dev.fitko.fitconnect.cli.commands.CreateTestKeysCommand;
import dev.fitko.fitconnect.cli.commands.GetAllSubmissionsCommand;
import dev.fitko.fitconnect.cli.commands.GetOneSubmissionCommand;
import dev.fitko.fitconnect.cli.commands.ListAllSubmissionsCommand;
import dev.fitko.fitconnect.cli.commands.SendBatchCommand;
import dev.fitko.fitconnect.cli.commands.SendSubmissionCommand;
import dev.fitko.fitconnect.cli.keygen.JWKGenerator;
import dev.fitko.fitconnect.cli.keygen.JWKPair;
import dev.fitko.fitconnect.cli.util.AttachmentDataType;
import dev.fitko.fitconnect.client.bootstrap.ApplicationConfigLoader;
import dev.fitko.fitconnect.client.bootstrap.ClientFactory;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static dev.fitko.fitconnect.cli.CommandLineRunner.DEFAULT_CONFIG_NAME;

class CommandExecutor {

    private static final Logger LOGGER = LoggerFactory.getLogger(CommandExecutor.class);

    private static final String IMPORTED_RECORD_TEMPLATE = "\n\n================== Imported record {} ==================\n";

    private final BatchImporter batchImporter;
    private final JWKGenerator jwkGenerator;
    private ApplicationConfig config;

    CommandExecutor(final BatchImporter batchImporter, final JWKGenerator jwkGenerator) {
        this.batchImporter = batchImporter;
        this.jwkGenerator = jwkGenerator;
    }

    void getOneSubmission(final GetOneSubmissionCommand getOneSubmissionCommand) throws IOException {

        LOGGER.info("Getting submission for id {}", getOneSubmissionCommand.submissionId);
        final var startTime = StopWatch.start();
        final var submission = ClientFactory.getSubscriberClient(loadConfig()).requestSubmission(getOneSubmissionCommand.submissionId);
        LOGGER.info("Submission download took {}", StopWatch.stopWithFormattedTime(startTime));
        if (submission == null) {
            LOGGER.info("No submission found for submission id {}", getOneSubmissionCommand.submissionId);
        } else {
            submission.acceptSubmission();
            writeData(submission, getTargetFolderPath(getOneSubmissionCommand));
        }
    }

    void getAllSubmissions(final GetAllSubmissionsCommand getAllSubmissionsCommand) throws IOException {
        final var destinationId = getAllSubmissionsCommand.destinationId;
        LOGGER.info("Getting all available submissions for destination {}", destinationId);
        final Set<SubmissionForPickup> submissions = ClientFactory.getSubscriberClient(loadConfig()).getAvailableSubmissionsForDestination(destinationId);
        for (final SubmissionForPickup submission : submissions) {
            final GetOneSubmissionCommand getOneSubmissionCommand = new GetOneSubmissionCommand();
            getOneSubmissionCommand.submissionId = submission.getSubmissionId();
            getOneSubmissionCommand.targetFolder = getAllSubmissionsCommand.targetFolder;
            getOneSubmission(getOneSubmissionCommand);
        }
    }

    void listSubmissions(final ListAllSubmissionsCommand listAllSubmissionsCommand) {
        final var destinationId = listAllSubmissionsCommand.destinationId;
        LOGGER.info("Listing available submissions for destination {}", destinationId);
        for (final SubmissionForPickup submission : ClientFactory.getSubscriberClient(loadConfig()).getAvailableSubmissionsForDestination(destinationId)) {
            LOGGER.info("caseId: {} - submissionId: {}", submission.getCaseId(), submission.getSubmissionId());
        }
    }

    SentSubmission sendSubmission(final SendSubmissionCommand sendSubmissionCommand) throws IOException {
        LOGGER.info("Sending new submission to destination {}", sendSubmissionCommand.destinationId);
        final Tika mimeTypeDetector = new Tika();
        final var startTime = StopWatch.start();
        final List<Attachment> attachments = sendSubmissionCommand.attachments.stream().map(toAttachment(mimeTypeDetector)).collect(Collectors.toList());
        final SentSubmission submission;
        if (sendSubmissionCommand.dataType == AttachmentDataType.json) {
            submission = sendWithJsonData(sendSubmissionCommand, attachments);
        } else {
            submission = sendWithXmlData(sendSubmissionCommand, attachments);
        }
        LOGGER.info("Import took {}", StopWatch.stopWithFormattedTime(startTime));
        return submission;
    }
    void sendBatch(final SendBatchCommand sendBatchCommand) throws BatchImportException, IOException {
        final List<ImportRecord> importRecords = batchImporter.readRecords(sendBatchCommand.dataPath);
        LOGGER.info("Sending batch of {} submissions", importRecords.size());
        var importCount = 0;
        final var startTime = StopWatch.start();
        for (final ImportRecord importRecord : importRecords) {
            if (sendSubmission(mapToSubmissionToSend(importRecord)) != null) {
                LOGGER.info(IMPORTED_RECORD_TEMPLATE, importCount + 1);
                importCount++;
            } else {
                LOGGER.info("Failed importing record {} => {}", importCount + 1, importRecord);
            }
        }
        LOGGER.info("DONE ! Finished batch import of {} submissions in {}", importCount, StopWatch.stopWithFormattedTime(startTime));
    }

    void createTestKeys(final CreateTestKeysCommand createTestKeysCommand) {

        final String dir = getDir(createTestKeysCommand.outputDir);
        final int keySize = createTestKeysCommand.keySize == 0 ? 4096 : createTestKeysCommand.keySize;

        final JWKPair encryptionKeyPair = jwkGenerator.generateEncryptionKeyPair(keySize );
        final JWKPair signatureKeyPair = jwkGenerator.generateSignatureKeyPair(keySize);

        writeFile(Path.of(dir, "publicKey_encryption.json"), "Encryption Public Key (key_use=wrapKey)", encryptionKeyPair.getPublicKey());
        writeFile(Path.of(dir, "privateKey_decryption.json"),"Decryption Private Key (key_use=unwrapKey)", encryptionKeyPair.getPrivateKey());

        writeFile(Path.of(dir, "publicKey_signature_verification.json"), "Signature Validation Public Key (key_use=verify)", signatureKeyPair.getPublicKey());
        writeFile(Path.of(dir, "privateKey_signing.json"), "Signing Private Key (key_use=sign)", signatureKeyPair.getPrivateKey());
    }

    private SentSubmission sendWithJsonData(final SendSubmissionCommand sendSubmissionCommand, final List<Attachment> attachments) throws IOException {

        final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                .setDestination(sendSubmissionCommand.destinationId)
                .setServiceType(sendSubmissionCommand.leikaKey, sendSubmissionCommand.serviceName)
                .setJsonData(getDataAsString(sendSubmissionCommand.data), URI.create(sendSubmissionCommand.schemaUri))
                .addAttachments(attachments)
                .build();

        return ClientFactory.getSenderClient(loadConfig()).send(sendableSubmission);
    }

    private SentSubmission sendWithXmlData(final SendSubmissionCommand sendSubmissionCommand, final List<Attachment> attachments) throws IOException {

        final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                .setDestination(sendSubmissionCommand.destinationId)
                .setServiceType(sendSubmissionCommand.leikaKey, sendSubmissionCommand.serviceName)
                .setXmlData(getDataAsString(sendSubmissionCommand.data), URI.create(sendSubmissionCommand.schemaUri))
                .addAttachments(attachments)
                .build();

        return ClientFactory.getSenderClient(loadConfig()).send(sendableSubmission);
    }

    private String getTargetFolderPath(final GetOneSubmissionCommand getOneSubmissionCommand) {
        if (getOneSubmissionCommand.targetFolder != null) {
            return getOneSubmissionCommand.targetFolder + "/" + getOneSubmissionCommand.submissionId;
        } else {
            return getOneSubmissionCommand.submissionId.toString();
        }
    }

    private String getDataAsString(final String dataPath) throws IOException {
        LOGGER.info("Reading data from {}", dataPath);
        return Files.readString(Path.of(dataPath));
    }

    private Function<String, Attachment> toAttachment(final Tika mimeTypeDetector) {
        return path -> Attachment.fromPath(Path.of(path), mimeTypeDetector.detect(path), Path.of(path).getFileName().toString(), "attachment");
    }

    private void writeData(final ReceivedSubmission receivedSubmission, final String dataDirPath) throws IOException {
        LOGGER.info("Creating data directory for submission in {}", dataDirPath);
        Files.createDirectories(Path.of(dataDirPath));

        final var fileEnding = AttachmentDataType.getFileTypeFromMimeType(receivedSubmission.getDataMimeType());
        final var filePath = Path.of(dataDirPath + "/data." + fileEnding);
        LOGGER.info("Writing data.{}", fileEnding);
        Files.write(filePath, receivedSubmission.getDataAsString().getBytes(StandardCharsets.UTF_8));

        for (final Attachment attachment : receivedSubmission.getAttachments()) {
            final String filename = attachment.getFileName();
            LOGGER.info("Writing attachment {}", filename);
            Files.write(Path.of(dataDirPath, filename), attachment.getDataAsBytes());
        }
    }

    private SendSubmissionCommand mapToSubmissionToSend(final ImportRecord importRecord) {
        final var submission = new SendSubmissionCommand();
        submission.setDestinationId(importRecord.getDestinationId());
        submission.setServiceName(importRecord.getServiceName());
        submission.setLeikaKey(importRecord.getLeikaKey());
        submission.setData(importRecord.getData().trim());
        submission.setDataType(importRecord.getDataType());
        submission.setSchemaUri(importRecord.getSubmissionDataSchemaUri());
        submission.setAttachments(Stream.of(importRecord.getAttachments().split(",")).map(String::trim).collect(Collectors.toList()));
        return submission;
    }

    private String getDir(final String outputDir) {
        if (outputDir != null) {
            return Path.of(outputDir).toAbsolutePath().toString();
        } else {
            return createTempDir();
        }
    }

    private static String createTempDir() {
        try {
            return Files.createTempDirectory("testJWKs").toFile().getAbsolutePath();
        } catch (final IOException e) {
            System.out.println(e);
            System.exit(0);
        }
        return null;
    }

    private static void writeFile(final Path path, final String name, final JWK jwk) {
        try {
            Files.write(path, toBytes(jwk));
            System.out.println("\nWrote " + name + " to \n" + path);
        } catch (final IOException e) {
            System.out.println(e);
            System.exit(0);
        }
    }

    private static byte[] toBytes(final JWK jwk) {
        return jwk.toJSONString().getBytes(StandardCharsets.UTF_8);
    }

    private ApplicationConfig loadConfig() {
        if(config != null) {
            return config;
        }
        try {
            return ApplicationConfigLoader.loadConfigFromEnvironment();
        } catch (final Exception e) {
            LOGGER.info("Loading config from default location {}", DEFAULT_CONFIG_NAME);
            return ApplicationConfigLoader.loadConfigFromPath(Path.of(DEFAULT_CONFIG_NAME));
        }
    }

}
