package dev.fitko.fitconnect.cli;

import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.cli.batch.CsvImporter;
import dev.fitko.fitconnect.cli.batch.ImportRecord;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.SendableEncryptedSubmission;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.nio.file.Path;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class CommandLineClientTest {

    private final LogCaptor logs = new LogCaptor();

    private CommandLineClient underTest;
    private SenderClient senderClientMock;
    private Subscriber subscriberMock;
    private SubscriberClient subscriberClientMock;

    @BeforeEach
    void setUp() {
        senderClientMock = mock(SenderClient.class);
        subscriberMock = mock(Subscriber.class);
        subscriberClientMock = mock(SubscriberClient.class);
        underTest = new CommandLineClient();
    }

    @Test
    void testInvalidDestinationId() {
        // When
        underTest.run("list", "--destinationId=123");

        // Then
        logs.assertContains("Parameter 123 is not a valid UUID.");
    }

    @Test
    void testListAllSubmissions() {

        // Given
        final var destinationId = UUID.randomUUID();
        final Set<SubmissionForPickup> submissions = generateSubmissions(destinationId, 10);
        when(subscriberClientMock.getAvailableSubmissionsForDestination(destinationId)).thenReturn(submissions);

        // When
        underTest.run("list", "--destinationId=" + destinationId);

        // Then
        logs.assertContains("Listing available submissions for destination " + destinationId);
        submissions.forEach(submission -> logs.assertContains(submission.getSubmissionId().toString()));
    }

    @Test
    void testGetSingleSubmission(@TempDir final Path tempDir) {

        // Given
        final var submissionId = UUID.randomUUID();

        final Attachment attachment = Attachment.fromString("attachment text", "text/plain", "testAttachment.txt", "test");

        final var expectedSubmission = new Submission();
        final var data = new ReceivedData("{ \"test\" : \"data\"", MimeType.APPLICATION_JSON);
        final var receivedSubmission = new ReceivedSubmission(subscriberMock, expectedSubmission, new Metadata(), data, List.of(attachment), Map.of(UUID.randomUUID(), "attachmentAuthTag"));

        when(subscriberClientMock.requestSubmission(submissionId)).thenReturn(receivedSubmission);

        // When
        underTest.run("get", "--submissionId=" + submissionId, "--target=" + tempDir);

        // Then
        logs.assertContains("Getting submission for id " + submissionId);
        logs.assertContains("Creating data directory for submission in " + tempDir);
        logs.assertContains("Writing data.json");
        logs.assertContains("Writing attachment testAttachment.txt");
    }

    @Test
    void testSendSubmission() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedSubmission = SentSubmission.builder()
                .submissionId(UUID.randomUUID())
                .build();

        when(senderClientMock.send(any(SendableSubmission.class))).thenReturn(expectedSubmission);

        // When
        underTest.run("send",
                "--destinationId=" + destinationId,
                "--serviceName=Test",
                "--leikaKey=urn:de:fim:leika:leistung:99400048079000",
                "--attachments=src/test/resources/attachment.txt",
                "--data=src/test/resources/data.json",
                "--dataType=json",
                "--schemaUri=https://schema.fitko.de/fim/test.schema.json");

        // Then
        logs.assertContains("Sending new submission to destination " + destinationId);
    }

    @Test
    void testBatchImport() {

        // Given

        final var testDataPath = "src/test/resources/batch_data.csv";
        final List<ImportRecord> importRecords = new CsvImporter().readRecords(testDataPath);

        when(senderClientMock.send(any(SendableEncryptedSubmission.class))).thenReturn(SentSubmission.builder().build());
        when(senderClientMock.send(any(SendableSubmission.class))).thenReturn(SentSubmission.builder().build());

        // When
        underTest.run("batch", "--data=" + testDataPath);

        // Then
        logs.assertContains("DONE ! Finished batch import of " + importRecords.size() + " submissions");
    }

    private Set<SubmissionForPickup> generateSubmissions(final UUID destinationId, final int count) {
        final Set<SubmissionForPickup> submissions = new HashSet<>(count);
        for (int i = 0; i < count; i++) {
            final SubmissionForPickup submission = new SubmissionForPickup();
            submission.setSubmissionId(UUID.randomUUID());
            submission.setDestinationId(destinationId);
            submissions.add(submission);
        }
        return submissions;
    }
}